
{-----------------------------------------------------------------------------
 Unit Name: Printing
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Routines for printing documents
-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}



unit Printing;

interface

uses
  Windows, Graphics;


function PreparePrint: Byte;
procedure PrintAll;

implementation

uses
  Math, Types, SysUtils, Classes,
  Forms, Controls, Printers,
  FParam4, ResourceStrings;

const
  cMinPageWidthPixels = 1000;

type
  StrBlock= record
    maxWidth: Integer;{max width in pixels of the strings in strs}
    strs: TStringList;
  end;

var
  bmp: TBitmap; {the bitmap where the colored background is drawn before sending
                 it to the printer}
  titleStrs: StrBlock;
  subStrs: array[1..5] of StrBlock;
  titleHeight, subStrHeight, lociMargin: Integer;
  r, rOr: TRect;

procedure ComputeLongFormula(var strs: StrBlock;
     can: TCanvas; theText: AnsiString;
     maxwidth, indentFirst, indentOthers: Integer);
{Adds to 'strs' the lines on which 'theText' is wrapped. 'indentFirst' is the number of
characters in which the first line is indented. 'indentOthers', the number of
chars for the indentation of the remaining lines. 'firstDesc' is the height in pixels
that must descend the first line. 'lineHeight' is the descent from line to line.}

const
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';

  kLeftArrow = 37;
  kUpArrow = 38;
  kDownArrow = 40;
  digits =['0','1','2','3','4','5','6','7','8','9'];
  specialchars = ['+','-','*','/','=',')',']','}','.',',',':',';', ' '];
var
  theLine, theBareText, firstHead, heads: AnsiString;
  theChar, auxChar: AnsiChar;
  index, i, textLength, b, previousLineEnd, charW, height, lineWidth: Integer;
begin
  textLength := Length(theText);
  theBareText:= '';
  strs.maxWidth:= 0;
  height:= can.Font.Height;
  if textLength > 0 then begin
    for index := 1 to textLength do begin
      theChar := theText[index];
      if not (theChar in [kReturn, kTab, kLineFeed]) then
        theBareText := theBareText + theChar
      else if length(theBareText) > 0 then begin
        auxChar:=theBareText[length(theBareText)];
        if (auxchar<>' ') then
          theBareText := theBareText + ' ';
      end;
    end;
    textLength := Length(theBareText);
  end;
  if textLength = 0 then
    Exit;

  firstHead:= '';
  if indentFirst > 0 then begin
    for i:= 1 to indentFirst do
      firstHead:= firstHead + ' ';
    theBareText:= firstHead + theBareText;
  end;
  lineWidth:= can.TextWidth(theBareText);
  if (lineWidth <= maxWidth) then begin
    strs.strs.Add(theBareText);
    strs.maxWidth:= lineWidth;
    Exit;
  end;

  textLength:= Length(theBareText);
  heads:= '';
  if indentOthers > 0 then
    for i:= 1 to indentOthers do
      heads:= heads+' ';

  b := 0;
  previousLineEnd:= 0;
  theLine:= firstHead;
  lineWidth:= can.TextWidth(theLine);
  index:= Max(0,indentFirst);
  while index < textLength do begin
    inc(index);
    theChar := theBareText[index];
    charW:= can.TextWidth(theChar);
    if (lineWidth + charW <= maxWidth) and (theChar in specialchars) and
           not((theChar= '.') and (index > 1) and (theBareText[index-1] in digits) and
              (index < textLength) and (theBareText[index+1] in digits)) then
      b:= index;
    lineWidth := lineWidth + charW;
    if index = textLength then begin
      theLine:= theLine + Copy(theBareText, previousLineEnd + 1, index - PreviousLineEnd);
      strs.strs.Add(theLine);

    end
    else  if (lineWidth > maxWidth) then begin
      if b <= previousLineEnd then begin
        theLine:= theLine + Copy(theBareText, previousLineEnd + 1, index -1 - PreviousLineEnd);
        strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
        strs.strs.Add(theLine);
        theLine:= heads;
        previousLineEnd:= index - 1;
      end
      else begin
        theChar:= theBareText[b];
        if theChar in ['+', '-', '*', '/', '=', ' '] then begin
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b -1 - PreviousLineEnd);
          strs.strs.Add(theLine);
          strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
          if theChar <> ' ' then
            theLine:= heads + theChar
          else
            theLine:= heads;
          previousLineEnd:= b;
        end
        else begin
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b - PreviousLineEnd);
          strs.strs.Add(theLine);
          strs.maxWidth:= max(strs.maxWidth, can.TextWidth(theLine));
          theLine:= heads;
          previousLineEnd:= b;
        end;
      end;
      b:= 0;
      lineWidth:= can.TextWidth(theLine);
      index:= previousLineEnd;
    end
  end;
end;

procedure PrintCurves(var can : Tcanvas; linesHeight: Integer; destRect: TRect);
const
  cMinDiv = 300;
  cDivStep = 40;
var
  i, w: Integer;
  pen : TPen;
begin
  pen := can.Pen;
  w := destRect.right - destRect.left;

  can.copyMode := cmsrcCopy;
  can.brush.color := clBlack;
  can.brush.style := bsSolid;

  can.Pen.Width:= max(1, w div 100);
  for i := kParabolicLocus to kZeroDiscrStrongPpalLocus do
    param4.DrawLocus(can, destRect, true, i);

  can.Pen.Width:= max(1, w div 500);
  param4.DrawLines(can, destRect);
  can.pen := pen;
end;

procedure GetTitleStrings(tx: AnsiString; maxWidth: Integer);
{Divide if necessary the title into several lines and store them
in 'titleStrs'}
const
  a1= 15.38 * 1.2;
  b1= 1.27 * 1.2;
  c= 1.3;
var
  can: TCanvas;
  fh: Single;
  i: Integer;
begin
  can:= Printer.Canvas;
  can.Font.Name := 'Helvetica';
  can.Font.Style := [fsBold];

  fh:= a1 + b1*sqrt(maxwidth - cMinPageWidthPixels); {set a size for the text
                       font in terms of 'maxWidth' and 'cMinPageWidthPixels'}
  can.Font.Height := round(fh);

  fh:= c*fh; {space between lines}

  titleStrs.strs.Clear;
  if (tx <> '') then
    ComputeLongFormula(titleStrs, can, tx, maxwidth,0,0);
  titleStrs.maxWidth:= 0;
  if titleStrs.strs.count > 0 then begin {compute title height and max width}
    titleHeight:= round((0.5 + titleStrs.strs.count)*fh);
    for i:= 0 to titleStrs.strs.count-1 do
      titleStrs.maxWidth:= Max(titleStrs.maxWidth, can.TextWidth(titleStrs.strs.Strings[i]));
  end
  else
    titleHeight:= 0;
end;



procedure GetSubTextStrings(tx: AnsiString; maxwidth: Integer);
{Divide if necessary each string of the caption into several lines and
 store them in 'subStrs'}
const
  a= 15.38;
  b= 1.27;
  c= 1.3;
var
  theText: AnsiString;
  can: TCanvas;
  shy: Single;
  i, n: Integer;
begin
  for i:= 1 to 5 do
    if Assigned(subStrs[i].strs) then
       subStrs[i].strs.Clear;
  can:= Printer.Canvas;
  can.Font.Name := 'Arial';
  can.Font.Style := [];
  shy:= a + b*sqrt(maxWidth - cMinPageWidthPixels);
  can.Font.Height:= round(shy);
  shy:= c*shy;

  if not param4.printEquations then begin
    if tx <> '' then
      ComputeLongFormula(subStrs[1], can, tx, maxwidth,3,0);
    Exit;
  end;

  theText := 'x(u,v) = ' + param4.inputStrs[1];
  ComputeLongFormula(subStrs[1], can, theText, maxwidth,0,3);

  theText := 'y(u,v) = ' + param4.inputStrs[2];
  ComputeLongFormula(subStrs[2], can, theText, maxwidth,0,3);

  theText := 'z(u,v) = ' + param4.inputStrs[3];
  ComputeLongFormula(subStrs[3], can, theText, maxwidth,0,3);

  theText := 'w(u,v) = ' + param4.inputStrs[4];
  ComputeLongFormula(subStrs[4], can, theText, maxwidth,0,3);

  theText := param4.inputStrs[5]+ ' < u < ' + param4.inputStrs[6]+ ' ,   '
             + param4.inputStrs[7]+ ' < v < ' + param4.inputStrs[8];
  subStrs[5].strs.Add(theText);
  subStrs[5].maxWidth:= can.TextWidth(theText);

  {compute total height of caption}
  n:= 0;
  for i:= 1 to 5 do
    n:= n+ subStrs[i].strs.Count;
  if n > 0 then
    subStrHeight:= round((n + 1.5)*shy);
end;

procedure PrintFrameAndTexts;
const
  a1= 15.38 * 1.2;
  b1= 1.27 * 1.2;

  a2= 15.38;
  b2= 1.27;

  c = 1.3;
var
  can: TCanvas;
  fh: Single;
  x, y, shy, i, j, w, pageW: Integer;
begin
  pageW:= Printer.pageWidth;

  {print title lines}
  if titleStrs.strs.count > 0 then begin
    can:= Printer.Canvas;
    can.Font.Name := 'Helvetica';
    can.Font.Style := [fsBold];
    fh:= a1 + b1*sqrt(r.Right-r.left - cMinPageWidthPixels);
    can.Font.Height := round(fh);
    shy:= round(c*fh); {shift of 'y' from a line to the next}
    y:= rOr.top - titleHeight;

    {center title}
    w:= titleStrs.maxWidth;
    x:= (pageW - w) div 2;
    for i:= 0 to titleStrs.strs.count-1 do begin
      can.TextOut(x,y,titleStrs.strs.Strings[i]);
      y:= y + shy;
    end;
  end;

  {frame the square where drawing will apear}
  Printer.Canvas.Pen.Width := 1;
  Printer.canvas.Pen.Color := clBlack;
  Printer.canvas.moveto(rOr.Left, rOr.Top);
  Printer.canvas.lineto(rOr.Left, rOr.Bottom);
  Printer.canvas.lineto(rOr.Right, rOr.Bottom);
  Printer.canvas.lineto(rOr.Right, rOr.Top);
  Printer.canvas.lineto(rOr.Left, rOr.Top);

  {print caption lines}
  can:= Printer.Canvas;
  can.Font.Name := 'Arial';
  can.Font.Style := [];
  fh:= a2 + b2*sqrt(r.right-r.left - cMinPageWidthPixels);
  can.Font.Height := round(fh);
  shy:= round(c*fh);
  y:= rOr.bottom + shy div 2 ;
  for j:=1 to 5 do
    if subStrs[j].strs.count > 0 then begin
      {center all lines}
      w:= subStrs[j].maxWidth;
      x:= (pageW - w) div 2;
      for i:= 0 to subStrs[j].strs.count-1 do begin
        can.TextOut(x,y,subStrs[j].strs.Strings[i]);
        y:= y + shy;
      end;
    end;
end;

procedure BitmapToPrinterCanvas(destRect: TRect; source: TBitmap);
{A routine that I have copied from somebody in the web that substitutes
 the crippled 'StrechtDraw' provided by Delphi}
var
  BitmapHeader: pBitmapInfo;
  BitmapImage: POINTER;
  HeaderSize: DWORD;
  ImageSize: DWORD;
  sh: HBitmap;
  ch: Hdc;
  w, h, srcw, srch, n: Integer;
  can: TCanvas;
begin
  can := Printer.Canvas;
  ch:= Printer.canvas.Handle;
  sh:= source.Handle;
  w := destRect.right - destRect.left;
  h := destRect.bottom - destRect.top;
  srcw:= source.Width;
  srch:= source.Height;
  GetDIBSizes(sh, HeaderSize, ImageSize);
  GetMem(BitmapHeader, HeaderSize);
  GetMem(BitmapImage, ImageSize);
  n:= 0;
  try
    GetDIB(sh, source.Palette, BitmapHeader^, BitmapImage^);
    n:= StretchDIBits(ch,
        destRect.left, destRect.top, // Destination Origin
        w, h, // Destination Width and Height
        0, 0, // Source Origin
        srcw, srch, // source Width and Height
        BitmapImage,
        TBitmapInfo(BitmapHeader^),
        DIB_RGB_COLORS,
        SRCCOPY);
  finally
    FreeMem(BitmapImage);
    FreeMem(BitmapHeader);
  end;
  if n=0 then
    can.StretchDraw(destRect,source);
end;

procedure GetPrinterParameters;
{establish margins and square from drawing}
var
  w: Integer;
  textsHeight: Integer;
begin
  textsHeight:= titleHeight + subStrHeight + lociMargin;
  w:= min( Printer.PageHeight*9 div 10 - textsHeight, Printer.PageWidth*9 div 10);
  r.left := (Printer.PageWidth -w) div 2;
  r.top := (Printer.PageHeight - w - textsHeight) div 2 + titleHeight;
  r.right := r.left + w;
  r.bottom := r.top + w;
  rOr := r;
  w:= w div 512;
  InflateRect(rOr, w, w);
end;

procedure ClipPrinter(r: TRect);
{limit the drawing in the printer to the rectangle 'r'}
  var
    rg: HRGN;
begin
  rg:= CreateRectRgnIndirect(r);
  ExtSelectClipRgn(Printer.Canvas.Handle,rg, RGN_AND);
  DeleteObject(rg);
end;

procedure RemoveClipPrinter;
{Make the clipping rectangle be the whole page}
    var
    rg: HRGN;
begin
  rg:= CreateRectRgn(0,0,Printer.PageWidth,Printer.PageHeight);
  SelectClipRgn(Printer.Canvas.Handle,rg);
  DeleteObject(rg);
end;

function PreparePrint: Byte;
var
  i, num, fineRes, prWidth: Integer;
begin
  result:= mrCancel;
  num := Printer.copies;
  if num < 1 then begin
    param4.printBackground:= False;
    param4.printBackground:= False;
    Exit;
  end;

  {set the width of the bitmap}
  prWidth:= (Printer.PageWidth * 9) div 10;
  fineRes:= Min(prWidth, param4.prResolution);
  printer.canvas.Brush.Color:= clWhite;
  printer.canvas.Brush.Style:= bsSolid;

  {create string lists for title and caption}
  if not Assigned(titleStrs.strs) then begin
    titleStrs.strs:= TStringList.Create;
    for i:= 1 to 5 do
      subStrs[i].strs:= TStringList.Create;
  end;
  try
    titleStrs.strs.Sorted:= False;
    for i:= 1 to 5 do
      subStrs[i].strs.sorted:= False;
  except on exception do begin
      param4.InfoBar.SimpleText:= eErrorInPrinting;
      param4.printBackground:= False;
      param4.printingBackground:= False;
      Exit;
    end;
  end;

  GetTitleStrings(param4.prTitle,prWidth);
  GetSubTextStrings(param4.prSubText, prWidth);
  GetPrinterParameters;
  if param4.printBackground and (param4.backgroundType >= 0) then begin
    {create 'bm'}
    if not Assigned(bmp) then
      bmp:= TBitmap.Create;
    try
      bmp.PixelFormat := pf32bit;
      bmp.Width := fineRes;
      bmp.Height := fineRes;
      param4.printingBackground:= True;

      {launch the threads that will do the drawing. After termination of
       all of them, 'PrintAll' will be called automatically}
      param4.ColorByFunction(bmp);
    except on Exception do begin
        bmp.free;
        PrintAll;
      end;
    end;
  end
  else
    PrintAll;
  result:= mrOK;
end;

procedure PrintAll;
{we write and draw everything to the printer}
var
  num, i: Integer;
  prCan: TCanvas;
begin
  num := Printer.copies;
  if num < 1 then
    Exit;
  prCan:= Printer.canvas;
  try
    for i := 1 to num do begin
      prCan.FillRect(rOr);
      RemoveClipPrinter;
      PrintFrameAndTexts;

      {clip printer canvas to the drawing rectangle}
      ClipPrinter(r);

      {print bitmap with color background}
      if param4.printBackground and (param4.backgroundType >= 0) and Assigned(bmp) then
        BitmapToPrinterCanvas(r, bmp);

      PrintCurves(prCan, printer.PageHeight, r);
      if (i < num) then
        Printer.NewPage;
    end;
  finally
    Printer.EndDoc;
    FreeAndNil(Bmp);
    param4.printBackground:= False;
    param4.printingBackground:= False;
  end;
end;



procedure Finalize;
var
  i: Integer;
begin
  titleStrs.strs.Free;
  for i:= 1 to 5 do
    subStrs[i].strs.Free;
  bmp.Free;
end;


initialization
  ;
finalization
  Finalize;


end.
 